import cv2
import base64
import numpy as np
from typing import List


class Input(object):
    def __init__(self, data):
        self.type = self.__type__()
        self.data = data

    def __type__(self) -> str:
        raise NotImplementedError(f'Method __type__() not implemented in {self.__name__}')


class ImageInput(Input):
    def __init__(self, base64_image: str):
        super().__init__(base64_image)

    def __type__(self):
        return 'image'

    @classmethod
    def from_cv2_image(cls, cv2_image):
        _, buffer = cv2.imencode('.jpg', cv2_image)
        return ImageInput(base64.b64encode(buffer).decode('utf-8'))

    @classmethod
    def from_pillow_image(cls, pillow_image):
        open_cv_image = np.array(pillow_image)
        cv2_image = open_cv_image[:, :, ::-1].copy()
        return cls.from_cv2_image(cv2_image)

    @classmethod
    def from_file(cls, file_path):
        return cls.from_cv2_image(cv2.imread(file_path))


class TextInput(Input):
    def __init__(self, text: str):
        super().__init__(text)

    def __type__(self):
        return 'text'


class GenericInput(Input):
    def __init__(self, data):
        super().__init__(data)

    def __type__(self) -> str:
        return 'generic'


class GenericArrayInput(GenericInput):
    def __init__(self, data: List):
        super().__init__(data)

    def __type__(self) -> str:
        return 'generic_array'


class NamedInput(Input):
    def __init__(self, name: str, input: Input):
        self.name = name
        self.input = input
        super().__init__(input.data)

    def __type__(self) -> str:
        return self.input.type


class CompositeInput(Input):
    def __init__(self, inputs: List[NamedInput]):
        result = {}
        for i in inputs:
            result[i.name] = i.input.data
        super().__init__(result)

    def __type__(self) -> str:
        return 'composite'

# TODO SoundInput(sound en formato estandar, metadata)



import logging
import os

from cryptography.hazmat.backends import default_backend as crypto_default_backend
from cryptography.hazmat.primitives import serialization as crypto_serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from flask import Flask, request, jsonify
from flask_cors import CORS, cross_origin
from halo import Halo
from waitress import serve

from starmart.__main__ import exit_after_seconds, bold


def server(on_result):
    app = Flask(__name__)
    CORS(app)
    app.config['CORS_HEADERS'] = 'Content-Type'

    log = logging.getLogger('werkzeug')
    log.setLevel(logging.ERROR)

    spinner = Halo(text='Waiting for browser authorization', spinner='dots')
    spinner.start()

    @app.route('/set-remote', methods=['POST'])
    @cross_origin()
    def set_remote():
        public_key = get_or_create_ssh_public_key()
        spinner.stop()
        on_result(request.json['remote'])
        return jsonify({'publicKey': public_key})

    @app.route('/set-clone', methods=['POST'])
    @cross_origin()
    def set_clone():
        spinner.stop()
        print('You already have an existing empty repository. Try calling',
              bold(f'starmart clone {request.json["repo_id"]}'))
        exit_after_seconds()
        return jsonify({'status': 'ok'})

    serve(app, host="127.0.0.1", port=4999)


def get_or_create_ssh_public_key():
    home = os.path.expanduser('~')
    ssh_dir = os.path.join(home, '.ssh')
    if not os.path.exists(ssh_dir):
        os.mkdir(ssh_dir)
        public_key = create_and_write_ssh_keypair()
    else:
        result = None
        find_file = False
        with open(os.path.join(home, '.ssh', 'config'), 'r') as f:
            for line in f:
                if line.startswith('gitlab.com'):
                    find_file = True
                elif find_file and line.startswith('IdentityFile ~/.ssh/'):
                    result = line.replace('IdentityFile ~/.ssh/', '')
                    break
        if result is None:
            public_key = create_and_write_ssh_keypair()
        else:
            public_key = ''
            with open(os.path.join(home, '.ssh', f'{result}.pub'), 'r') as f:
                for l in f:
                    public_key += l
                public_key = public_key.replace('\n', '')

    return public_key


def create_and_write_ssh_keypair():
    key = rsa.generate_private_key(
        backend=crypto_default_backend(),
        public_exponent=65537,
        key_size=2048
    )

    private_key = key.private_bytes(
        crypto_serialization.Encoding.PEM,
        crypto_serialization.PrivateFormat.PKCS8,
        crypto_serialization.NoEncryption()
    )

    public_key = key.public_key().public_bytes(
        crypto_serialization.Encoding.OpenSSH,
        crypto_serialization.PublicFormat.OpenSSH
    )

    home = os.path.expanduser('~')
    ssh_dir = os.path.join(home, '.ssh')

    with open(os.path.join(ssh_dir, 'gitlab'), 'wb') as f:
        f.write(private_key)

    with open(os.path.join(ssh_dir, 'gitlab.pub'), 'wb') as f:
        f.write(public_key)

    with open(os.path.join(home, '.ssh', 'config'), 'a') as f:
        f.write(f'Host gitlab.com\n\tHostName gitlab.com\n\tIdentityFile ~/.ssh/gitlab\n')

    return public_key.decode('utf-8')



import _thread
import argparse
import time
import webbrowser
from threading import Thread

from git import Repo, InvalidGitRepositoryError
from halo import Halo

from starmart.config.config import Config


class Action(object):
    def __init__(self, args):
        self.config = Config.default_config()
        self.args = args

    def act(self):
        raise NotImplementedError(f'act not implemented in {self.__name__}')

    @classmethod
    def get_action(cls):
        actions = dict({
            'deploy': DeployAction,
            'init': InitAction,
            'clone': CloneAction
        })

        args = cls.__parse_arguments__()
        action = actions.get(args.action[0])
        if action is None:
            raise ValueError('Action should be deploy, init or clone')

        return action(args)

    @classmethod
    def __parse_arguments__(cls):
        # configuring arguments
        parser = argparse.ArgumentParser()
        parser.add_argument('action', nargs=1, type=str, default='None',
                            help='Run init on a new project, deploy to push the code or clone <project_id> to retrieve an existing project')
        parser.add_argument('project_id', nargs='?', help='The project id', default=None)
        return parser.parse_args()


class InitAction(Action):

    def act(self):
        self.__auth_with_web_browser__()

    @Halo(text='Cloning starter code repo', spinner='dots')
    def __clone_default_code__(self):
        cloned = Repo.clone_from(self.config.github_repo(), 'starter_code')
        for r in cloned.remotes:
            if r.name == 'origin':
                cloned.delete_remote(r)
                break
        return cloned

    def __auth_with_web_browser__(self):
        webbrowser.open(f'{self.config.authentication_host()}/development/login')

        def callback(url):
            remote_host = self.config.git_remote_host()
            if not url.startswith(remote_host):
                raise ValueError(f'URL does not match the authentication host: {remote_host}')
            repo = self.__clone_default_code__()
            repo.create_remote('starmart', url=url)
            print('Happy coding!')

            # this is needed to exit flask server -> first it needs to return and then exit
            exit_after_seconds()

        # this blocks because of the server. that's why I set a callback
        from starmart.server.Server import server
        server(callback)


class DeployAction(Action):
    def act(self):
        self.__configure_repo__()

    @Halo(text='Pushing latest commits', spinner='dots')
    def __configure_repo__(self):
        try:
            repo = Repo('.')
            remote = None
            for r in repo.remotes:
                if r.name == 'starmart':
                    remote = r
                    break
            if remote is None:
                raise ValueError(f'The repository does not contain the starmart remote. Please call' +
                                 f' {bold("starmart init")}, before calling {bold("starmart deploy")}.')
            remote.push()
            print('\nPushed. Happy coding!')
        except InvalidGitRepositoryError:
            raise ValueError('Github repository not initialized. Call starmart init before calling starmart deploy.')


class CloneAction(Action):

    def act(self):
        self.__clone_repo__()

    def __clone_repo__(self):
        project_id = self.args.project_id
        if project_id is None:
            raise ValueError(bold('starmart clone') + ' needs the project id')
        spinner = Halo(text=f'Cloning project {project_id}', spinner='dots')
        spinner.start()
        repo = Repo.clone_from(f'{self.config.git_remote_host()}/{project_id}.git', f'starmart_project_{project_id}')
        repo.remote('origin').rename('starmart')
        spinner.stop()
        print('Cloned. Happy coding!')


def exit_after_seconds(seconds=2):
    def do_exit():
        time.sleep(seconds)
        _thread.interrupt_main()

    Thread(target=do_exit).start()


def bold(text):
    return '\033[1m' + text + '\033[0m'


def main():
    Action.get_action().act()


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        exit(0)



from distutils.core import setup

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='starmart',
    packages=['starmart', 'starmart.server', 'starmart.config'],
    version='0.1.9',
    license='apache-2.0',
    description='Starmart deployment tool',
    author='Tomas Piaggio',
    author_email='tomaspiaggio@starmart.io',
    url='https://starmart.io',
    keywords=['AI', 'Machine Learning', 'Deep Learning', 'Serverless'],
    install_requires=[
        'GitPython==3.1.24',
        'flask==2.0.2',
        'waitress==2.0.0',
        'halo==0.0.31',
        'cryptography==36.0.1',
        'opencv-python~=4.5.3',
        'numpy==1.21.2',
        'flask-cors==3.0.10'
    ],
    scripts=['bin/starmart'],
    long_description=long_description,
    long_description_content_type='text/markdown'
)



class Config(object):
    def github_repo(self) -> str:
        raise NotImplementedError(f'github_repo not implemented in {self.__name__}')

    def authentication_host(self) -> str:
        raise NotImplementedError(f'authentication_host not implemented in {self.__name__}')

    def git_remote_host(self):
        raise NotImplementedError(f'git_remote_host not implemented in {self.__name__}')

    @classmethod
    def default_config(cls) -> 'Config':
        try:
            from starmart.config.development_config import DevelopmentConfig
            return DevelopmentConfig()
        except ModuleNotFoundError:
            from starmart.config.production_config import ProductionConfig
            return ProductionConfig()
